/*
    Notes

    - Since the uPD7759 can only address up to 128k of ROM, some address
      (maybe one of the I/O chip ports) might control the upper two address
      lines - the largest sample data ROMs are 512k. (e.g. for Puyo Puyo 2
      each 128k chunk of the ROM has it's own header)

    - The timing is off. The 68000 runs at 10MHz, currently it is running
      at about 7.67MHz (same as NTSC Genesis). This means the H counter
      is being incorrectly calculated, though I don't know of any games
      that use it.

    The level 6 interrupt is always enabled, regardless of the setting for
    bit 5 in register #1. Perhaps the bit was moved or another interrupt
    enable register exists somewhere.

    The I/O chip (315-5296) is connected to the 68000 through A0-5
    and D0-D8. Based on the System 24 schematics, the address bus
    is laid out like so:

    I/O    68000
    ------------
    A0 <-> A1
    A1 <-> A2
    A2 <-> A3
    A3 <-> A4
    A4 <-> A5
    A5 <-> A8

    So the I/O chip occupies *odd* addresses through 840000-84001F and
    840100-84011F, rather than just 840000-84003F.

    These addresses correspond to ports A through H.

    $840001 : Player 1 input
    $840003 : Player 2 input
    $840005 : Unknown (Bit 7 tested during gameplay)
    $840007 : Unknown
    $840009 : Control panel
    $84000B : DIP switch #1
    $84000D : DIP switch #2
    $84000F : Bits 1-0 select the palette RAM mapped at 8C0000-8C03FF
              Purpose of other bits are unknown.

    Not sure how these are mapped. Internal registers?
    Bits 2-0 of $84001D may be for pins CNT0-2.

    $840011 : Poto Poto protection ('S')
    $840013 : Poto Poto protection ('E')
    $840015 : Poto Poto protection ('G')
    $840017 : Poto Poto protection ('A')
    $840019 : Unknown
    $84001B : Unknown
    $84001D : Unknown
    $84001F : Unknown ($88 written to at startup)

    Probably A5 also causes /FMCS to be asserted.

    $840101 : YM3438 status / register latch (Ch.1-3)
    $840103 : YM3438 register data (Ch.1-3)
    $840105 : YM3438 register latch (Ch.4-6)
    $840107 : YM3438 register data (Ch.4-6)

    The remaining addresses for when A5 = 1 are unused.
    Maybe the YM3438 addresses are mirrored up to $84011F.

    $840005, $840007, $84001D, are read and written to.
    Could have something to do with an interrupt enable, or controlling the
    reset line for the YM3438.

    Other unknown addresses that are probably unrelated to the I/O chip:

    $880335 - Not accessed much, but always written to four bytes at a time.
              tfacbl does the same thing to $880135, could be a mirror.
    $800201 - Has values 06 and 07 frequently written.
*/


#include "shared.h"

byte rom[ROM_SIZE];
byte pcm[PCM_SIZE];
byte ram[0x10000];
static byte status_toggle = 0;  /* Hack to replace YM3438 timer emulation */

byte masterpalram[0x1000];
byte palram[0x1000];
byte pal_bank       = 0;
byte pal_select     = 0;
int palmode=0;
int gab = 1;
int gab2 = 1;
int palbank2=0;


void genesis_frame(int skip)
{
    int line;

    vdp->status = (STATUS_D13 | STATUS_D12 | STATUS_D10 | STATUS_FIFOEMPTY);
    for(line = 0; line < 262; line += 1)
    {
        vdp->curline = line;

        m68k_clear_cycles();
        m68k_execute(512);

        if(line < 0xE0)
        {
            if(!line) vdp->counter = vdp->reg[10];
            if(vdp->counter)
            {
                vdp->counter -= 1;
            }
            else
            {
                vdp->counter = vdp->reg[10];
                if(vdp->reg[0] & 0x10)
                {
                    m68k_cause_interrupt(4);
                }
            }

              vdp->status |= STATUS_HBLANK;
              m68k_execute(64);
              vdp->status &= STATUS_HBLANK;

            if((skip == 0) && (line < 224))
            {
                render_line(line);
            }
        }

        if(line == 0xE0)
        {
            vdp->status |= STATUS_VBLANK;
            vdp->status |= STATUS_VINT;
            m68k_cause_interrupt(6);
        }
    }
}


void genesis_init(void)
{
    /* Do game-specific initialization, if used */
    if(game_driver->init) game_driver->init();

    /* Clear RAM */
    memset(ram, 0, 0x10000);

    /* This is specific to Starscream */
    byteswap(rom, ROM_SIZE);

    /* Initialize Starscream and switch to the first CPU */
    m68k_init();
    m68k_switch(0);

    /* ROM */
    m68k_add_fetch     (0x000000, 0x1FFFFF, (unsigned)rom);
    m68k_add_read      (0x000000, 0x1FFFFF, NULL, rom);

    /* RAM */
    m68k_add_fetch     (0xFF0000, 0xFFFFFF, (unsigned)ram);
    m68k_add_readwrite (0xFF0000, 0xFFFFFF, NULL, ram);

    /* VDP */
    m68k_add_readbyte  (0xC00000, 0xC0001F, vdp_r_8, NULL);
    m68k_add_writebyte (0xC00000, 0xC0001F, vdp_w_8, NULL);
    m68k_add_readword  (0xC00000, 0xC0001F, vdp_r_16, NULL);
    m68k_add_writeword (0xC00000, 0xC0001F, vdp_w_16, NULL);

    /* Palette RAM */
    m68k_add_readwrite (0x8C0000, 0x8C0FFF, NULL, palram);

    /* Puyo Puyo 2 accesses mirrored RAM */
    m68k_add_fetch     (0xFE0000, 0xFEFFFF, (unsigned)ram);
    m68k_add_readwrite (0xFE0000, 0xFEFFFF, NULL, ram);

    /* Debugging - trap all unmapped/unhandled writes */
    m68k_add_readbyte  (0x000000, 0xFFFFFF, m68k_read_memory_8, NULL);
    m68k_add_readword  (0x000000, 0xFFFFFF, m68k_read_memory_16, NULL);
    m68k_add_writebyte (0x000000, 0xFFFFFF, m68k_write_memory_8, NULL);
    m68k_add_writeword (0x000000, 0xFFFFFF, m68k_write_memory_16, NULL);

    /* Reset the 68000 */
    m68k_reset();

    /* Log initial PC and stack top */
    error("PC:%08X\tSP:%08X\n", m68k_readpc(), s68000context.areg[7]);
}


void genesis_trash(void)
{
    /* Log PC and stack top on exit */
    error("PC:%08X\tSP:%08X\n", m68k_readpc(), s68000context.areg[7]);
}


void genesis_reset(void)
{
    memset(ram, 0, 0x10000);
    status_toggle = 0;
    m68k_reset();
}


/*--------------------------------------------------------------------------*/
/* Memory handlers                                                          */
/*--------------------------------------------------------------------------*/

unsigned int m68k_read_memory_8(unsigned int address)
{
    byte temp = 0xFF;

    switch(address)
    {
        case 0x840001: /* Player 1 input */
            if(input.pad[0] & INPUT_LEFT ) temp &= ~0x80;
            if(input.pad[0] & INPUT_RIGHT) temp &= ~0x40;
            if(input.pad[0] & INPUT_UP   ) temp &= ~0x20;
            if(input.pad[0] & INPUT_DOWN ) temp &= ~0x10;
            if(input.pad[0] & INPUT_A    ) temp &= ~0x04;
            if(input.pad[0] & INPUT_B    ) temp &= ~0x02;
            if(input.pad[0] & INPUT_C    ) temp &= ~0x01;
            return (temp);

        case 0x840003: /* Player 2 input */
            if(input.pad[1] & INPUT_LEFT ) temp &= ~0x80;
            if(input.pad[1] & INPUT_RIGHT) temp &= ~0x40;
            if(input.pad[1] & INPUT_UP   ) temp &= ~0x20;
            if(input.pad[1] & INPUT_DOWN ) temp &= ~0x10;
            if(input.pad[1] & INPUT_A    ) temp &= ~0x04;
            if(input.pad[1] & INPUT_B    ) temp &= ~0x02;
            if(input.pad[1] & INPUT_C    ) temp &= ~0x01;
            return (temp);

        case 0x840009: /* Control panel */
            if(input.system & INPUT_1PCOIN ) temp &= ~0x01;
            if(input.system & INPUT_2PCOIN ) temp &= ~0x02;
            if(input.system & INPUT_1PSTART) temp &= ~0x10;
            if(input.system & INPUT_2PSTART) temp &= ~0x20;
            if(input.system & INPUT_SERVICE) temp &= ~0x08;
            if(input.system & INPUT_TEST   ) temp &= ~0x04;
            return (temp);

        case 0x84000B:  /* DIP switch #1 */
            return (dip_switch[0]);

        case 0x84000D:  /* DIP switch #2 */
            return (dip_switch[1]);

        /* PotoPoto protection */
        case 0x840011: return ('S');
        case 0x840013: return ('E');
        case 0x840015: return ('G');
        case 0x840017: return ('A');

        case 0x840101:  /* YM3438 status */
            status_toggle ^= 0xFF;
            return (status_toggle);

        case 0x800201: /* Unknown */
            return (temp);

        default:
            error("read byte from %08X (%08X)\n", address, m68k_readpc());
            break;
    }

    return (0xFF);
}


void m68k_write_memory_8(unsigned int address, unsigned int value)
{
    switch(address)
    {
        case 0x800001: /* Protection device & palette select */
            pal_select = (value & 0x0F);
            palbank2 = value;
            break;

        case 0x84000F: /* Seems to be Palette banking */
		/* I have a feeling theres probably a better way of doing this .. but this was quick to set up */
        if (gab == 1)
		{
            value &= 3;
            memcpy(&masterpalram[pal_bank*0x400], palram, 0x400);  // Back up Current Palette Bank
            pal_bank = value & 3; // Get New Palette Bank Value
            memcpy(palram, &masterpalram[pal_bank*0x400], 0x400);  // Restore that New Bank from RAM
		}
		break;

        case 0x880001: /* uPD7759 message_w */
            if(snd.enabled)
            {
                UPD7759_reset_w(0, 0);
                UPD7759_message_w(0, value);
                UPD7759_start_w(0, 0);
            }
            break;
                       /* YM3438 ports */
        case 0x840101: if(snd.enabled) YM2612Write(0, 0, value); break;
        case 0x840103: if(snd.enabled) YM2612Write(0, 1, value); break;
        case 0x840105: if(snd.enabled) YM2612Write(0, 2, value); break;
        case 0x840107: if(snd.enabled) YM2612Write(0, 3, value); break;

        case 0x800201: /* Unknown */
            break;

        default:
            error("write byte %02X to %08X (%08X)\n", value, address, m68k_readpc());
            break;
    }
}

unsigned int m68k_read_memory_16(unsigned int address)
{
    error("read word from %08X (%08X)\n", address, m68k_readpc());
    return (0xFFFF);
}

void m68k_write_memory_16(unsigned int address, unsigned int value)
{
    error("write word %04X to %08X (%08X)\n", value, address, m68k_readpc());
}


