
#include "osd.h"
#include "uifont.h"
#include "m13_pal.h"

int ui_colors[0x40];
PALETTE gen_pal;
FILE *errorlog                  = NULL;
BITMAP *gen_bmp                 = NULL;
volatile int frame_count        = 0;
volatile int frame_rate         = 0;
volatile int tick_count         = 0;
volatile int old_tick_count     = 0;
int skip                        = 0;
int quit                        = 0;
volatile int msg_enable         = 0;
char msg[0x100];
int sound_enable                = -1;   // 0=fm 1=psg 2=upd

int main (int argc, char **argv)
{
    /* Default message if no game specified */
    if(argc < 2) {
        printf("c2emu - Sega System C2 Arcade Game Emulator\n");
        printf("(C) 1999, 2000  Charles Mac Donald\n");
        printf("Version 0.2, build date: %s, %s\n", __DATE__, __TIME__);
        printf("Usage: c2emu <gamename> [-options]\n\n");
        printf("Type 'c2emu -help' for a summary of the available options\n");
        printf("Type 'c2emu -list' for a list of game drivers.\n");
        printf("Type 'c2emu -info <gamename>' for a list of files needed by game drivers.\n");
        exit(1);
    }

    /* List available options */
    if(stricmp(argv[1], "-help") == 0) { print_options(); exit(1); };
    if(stricmp(argv[1], "-list") == 0) { print_driver_list(); exit(1); };
    if(stricmp(argv[1], "-info") == 0) { print_driver_files(argv[2]); exit(1); };

#if DEBUG
    errorlog = fopen("error.log","w");
#endif

    /* Handle system configuration */
    do_config("c2emu.cfg");

    if(!load_driver(argv[1]))
    {
        printf("Error loading `%s'\n", argv[1]);
        exit(1);
    }

    init_machine();
    system_init(option.sndrate);           

    while(!quit)
    {
        frame_count += 1;
        get_input();
        skip = (frame_count % option.skip == 0) ? 0 : 1;
        genesis_frame(skip);
        dos_update_audio();
        if(!skip) {
            if(option.fps) msg_print(2, 2, "%d", frame_rate);
            if(msg_enable) msg_print(4, BITMAP_HEIGHT - 12, "%s", msg);
            dos_update_video();
        }
    }

    free_driver();
    system_shutdown();
    trash_machine();
    if(errorlog) fclose(errorlog);
    return (0);
}


void tick_handler(void)
{
    tick_count += 1;
    if(msg_enable) --msg_enable;
    if(tick_count % 60 == 0)
    {
        frame_rate = frame_count;
        frame_count = 0;
    }
}
END_OF_FUNCTION(tick_handler)

void msg_print(int x, int y, char *fmt, ...)
{
    int i = 32;
    int j = 0;
    va_list ap;
    char token[0x100];
    char str[0x100];

    strcpy(str, "\0");
    va_start(ap, fmt);
    vsprintf(token, fmt, ap);
    strcat(str, token);
    va_end(ap);

    textprintf(gen_bmp, font, i+x+1, j+y  , COLOR_BLACK, "%s", str);
    textprintf(gen_bmp, font, i+x-1, j+y  , COLOR_BLACK, "%s", str);
    textprintf(gen_bmp, font, i+x  , j+y+1, COLOR_BLACK, "%s", str);
    textprintf(gen_bmp, font, i+x  , j+y-1, COLOR_BLACK, "%s", str);
    textprintf(gen_bmp, font, i+x  , j+y  , COLOR_WHITE, "%s", str);
}


void add_msg(char *fmt, ...)
{
    va_list ap;
    char token[0x100];

    msg_enable = 60;
    strcpy(msg, "\0");
    va_start(ap, fmt);
    vsprintf(token, fmt, ap);
    strcat(msg, token);
    va_end(ap);
}



void init_machine(void)
{
    int ret;

    if(option.sound)
        msdos_init_sound(&option.sndrate, option.sndcard);

    allegro_init();

    set_color_depth(option.video_depth);
    if(option.scanlines) option.video_height *= 2;
    ret = set_gfx_mode(option.video_driver, option.video_width, option.video_height, 0, 0);
    if(ret != 0) {
        printf("Error setting graphics mode (%dx%d %dbpp).\nAllegro says: `%s'\n", option.video_width, option.video_height, option.video_depth, allegro_error);
        exit(1);
    }
    clear(screen);

    if(option.video_depth == 8)
    {
        COLOR_WHITE = 0xFE;
        COLOR_BLACK = 0xFF;

        set_palette(m13_pal);

        inp(0x3DA);
        outp(0x3C0, 0x31);
        outp(0x3C0, COLOR_BLACK);

        clear_to_color(screen, COLOR_BLACK);
    }
    else
    {
        COLOR_BLACK = makecol(0, 0, 0);
        COLOR_WHITE = makecol(0xFF, 0xFF, 0xFF);
    }

    text_mode(-1);
    font = &zyrinx_font;
    gen_bmp = create_bitmap(BITMAP_WIDTH, BITMAP_HEIGHT);
    clear_to_color(gen_bmp, COLOR_BLACK);

    memset(&bitmap, 0, sizeof(t_bitmap));
    bitmap.data = (byte *)&gen_bmp->line[0][0];
    bitmap.width = BITMAP_WIDTH;
    bitmap.height = BITMAP_HEIGHT;
    bitmap.depth = option.video_depth;
    bitmap.pitch = (bitmap.depth >> 3) * bitmap.width;

    install_keyboard();
    install_joystick(option.joy_driver);

    LOCK_VARIABLE(frame_count);
    LOCK_VARIABLE(frame_rate);
    LOCK_VARIABLE(tick_count);
    LOCK_VARIABLE(msg_enable);
    LOCK_FUNCTION(tick_handler);
    install_int_ex(tick_handler, BPS_TO_TIMER(60));
}


void trash_machine(void)
{
    if(option.sound) msdos_shutdown_sound();
    clear(screen);
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
}


void get_input(void)
{
    /* Clear input */
    input.pad[0] = 0;
    input.pad[1] = 0;
    input.system = 0;

    /* Check Player 1 keys */
    if(key[KEY_UP])    input.pad[0] |= INPUT_UP;
    else
    if(key[KEY_DOWN])  input.pad[0] |= INPUT_DOWN;
    if(key[KEY_LEFT])  input.pad[0] |= INPUT_LEFT;
    else
    if(key[KEY_RIGHT]) input.pad[0] |= INPUT_RIGHT;
    if(key[KEY_A])     input.pad[0] |= INPUT_A;
    if(key[KEY_S])     input.pad[0] |= INPUT_B;
    if(key[KEY_D])     input.pad[0] |= INPUT_C;

    /* Sound channel keys */                
    if(check_key(KEY_7)) { sound_enable ^= 1; add_msg("YM3438 sound %s", sound_enable & 1 ? "enabled" : "disabled"); };
    if(check_key(KEY_8)) { sound_enable ^= 2; add_msg("SN76489 sound %s", sound_enable & 2 ? "enabled" : "disabled"); };
    if(check_key(KEY_9)) { sound_enable ^= 4; add_msg("uPD7759 sound %s", sound_enable & 4 ? "enabled" : "disabled"); };
    if(check_key(KEY_0)) { sound_enable = -1; add_msg("All sound channels enabled"); };

    /* Control panel keys */
    if(key[KEY_1])     input.system |= INPUT_1PSTART;
    if(key[KEY_2])     input.system |= INPUT_2PSTART;
    if(key[KEY_3])     input.system |= INPUT_1PCOIN;
    if(key[KEY_4])     input.system |= INPUT_2PCOIN;
    if(key[KEY_5])     { input.system |= INPUT_SERVICE; add_msg("Service"); };
    if(key[KEY_6])     { input.system |= INPUT_TEST;    add_msg("Test"); };

    /* Exit emulator */
    if(key[KEY_END] || key[KEY_ESC])  quit = 1;

    /* Reset virtual console */
    if(check_key(KEY_TAB))  { system_reset(); add_msg("Reset"); };

    /* Frame skip control */
    if(check_key(KEY_F1)) { option.skip = 1; add_msg("Frame skip disabled"); };
    if(check_key(KEY_F2)) { option.skip = 2; add_msg("Frame skip set to 1"); };
    if(check_key(KEY_F3)) { option.skip = 3; add_msg("Frame skip set to 2"); };
    if(check_key(KEY_F4)) { option.skip = 4; add_msg("Frame skip set to 3"); };

    /* Save a PCX snapshot */
    if(check_key(KEY_F5))
    {
        int ret = 0;
        static int snap_count = 0;
        char path[0x100];
        BITMAP *temp = NULL;
        PALETTE pal;
        get_palette(pal);
        temp = create_bitmap(vp_width, vp_height);
        blit(gen_bmp, temp, BITMAP_MARGIN, 0, 0, 0, vp_width, vp_height);
        sprintf(path, "%s/%s-%d.pcx", dirs.snap, game_driver->drvname, snap_count++);
        ret = save_pcx(path, temp, pal);
        destroy_bitmap(temp);
        if(!ret) add_msg("Saved snapshot");
    }

    /* DIP switch bank select and switch toggle */
    if(check_key(KEY_F6))    { dip_used=1; dip_bank ^= 1;                 add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };
    if(check_key(KEY_1_PAD)) { dip_used=1; dip_switch[dip_bank] ^= 0x80;  add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };
    if(check_key(KEY_2_PAD)) { dip_used=1; dip_switch[dip_bank] ^= 0x40;  add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };
    if(check_key(KEY_3_PAD)) { dip_used=1; dip_switch[dip_bank] ^= 0x20;  add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };
    if(check_key(KEY_4_PAD)) { dip_used=1; dip_switch[dip_bank] ^= 0x10;  add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };
    if(check_key(KEY_5_PAD)) { dip_used=1; dip_switch[dip_bank] ^= 0x08;  add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };
    if(check_key(KEY_6_PAD)) { dip_used=1; dip_switch[dip_bank] ^= 0x04;  add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };
    if(check_key(KEY_7_PAD)) { dip_used=1; dip_switch[dip_bank] ^= 0x02;  add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };
    if(check_key(KEY_8_PAD)) { dip_used=1; dip_switch[dip_bank] ^= 0x01;  add_msg("DIP#%d = %d%d%d%d%d%d%d%d", dip_bank, (dip_switch[dip_bank]>>7)&1,(dip_switch[dip_bank]>>6)&1,(dip_switch[dip_bank]>>5)&1,(dip_switch[dip_bank]>>4)&1,(dip_switch[dip_bank]>>3)&1,(dip_switch[dip_bank]>>2)&1,(dip_switch[dip_bank]>>1)&1,(dip_switch[dip_bank]>>0)&1); };

	/* Speed control */
    if(check_key(KEY_F9)) { option.fps ^= 1; add_msg("FPS meter %s", option.fps ? "on" : "off"); };
    if(check_key(KEY_F10)) { option.throttle ^= 1; add_msg("Speed throttle %s", option.throttle ? "on" : "off"); };
    if(check_key(KEY_F11)) { option.vsync ^= 1; add_msg("VSync %s", option.vsync ? "on" : "off"); };

    /* Palette mode select */
    if(check_key(KEY_F12))
    {
        palmode ^= 1;
        if(palmode == 0) set_palette(m13_pal);
        add_msg("Using %s palette", palmode ? "Dynamic" : "Fixed");
    }

    /* Alt-R dumps ROM for debugging */
    if(key[KEY_R] && key[KEY_ALTGR])
    {
        char path[0x100];
        sprintf(path, "%s.rom", game_driver->drvname);
        byteswap(rom, ROM_SIZE);
        save_file(path, rom, ROM_SIZE);
        byteswap(rom, ROM_SIZE);
        add_msg("Dumped ROM to %s", path);
    }

    /* Is the joystick being used ? */
    if(option.joy_driver != JOY_TYPE_NONE)
    {
        poll_joystick();

        /* Check player 1 joystick */
        if(joy[0].stick[0].axis[1].d1) input.pad[0] |= INPUT_UP;
        else
        if(joy[0].stick[0].axis[1].d2) input.pad[0] |= INPUT_DOWN;

        if(joy[0].stick[0].axis[0].d1) input.pad[0] |= INPUT_LEFT;
        else
        if(joy[0].stick[0].axis[0].d2) input.pad[0] |= INPUT_RIGHT;

        if(joy[0].button[0].b)  input.pad[0] |= INPUT_A;
        if(joy[0].button[1].b)  input.pad[0] |= INPUT_B;
        if(joy[0].button[2].b)  input.pad[0] |= INPUT_C;

        /* More than one joystick supported ? */
        if(num_joysticks > 2)
        {
            /* Check player 2 joystick */
            if(joy[1].stick[0].axis[1].d1) input.pad[1] |= INPUT_UP;
            else
            if(joy[1].stick[0].axis[1].d2) input.pad[1] |= INPUT_DOWN;

            if(joy[1].stick[0].axis[0].d1) input.pad[1] |= INPUT_LEFT;
            else
            if(joy[1].stick[0].axis[0].d1) input.pad[1] |= INPUT_RIGHT;

            if(joy[1].button[0].b)  input.pad[1] |= INPUT_A;
            if(joy[1].button[1].b)  input.pad[1] |= INPUT_B;
            if(joy[1].button[2].b)  input.pad[1] |= INPUT_C;
        }
    }
}

/*--------------------------------------------------------------------------*/
/* Helper functions                                                         */
/*--------------------------------------------------------------------------*/

int check_key(int code)
{
    static char lastbuf[0x100] = {0};

    if((!key[code]) && (lastbuf[code] == 1))
        lastbuf[code] = 0;

    if((key[code]) && (lastbuf[code] == 0))
    {
        lastbuf[code] = 1;
        return (1);
    }                                                                    

    return (0);
}

void error(char *format, ...)
{
#if DEBUG
    va_list ap;
    va_start(ap, format);
    vfprintf(errorlog, format, ap);
    va_end(ap);
#endif
}

int load_file_even(char *filename, char *mem, int size)
{
    char *buf;
    int i, file_size;
    FILE *fd = NULL;
    fd = fopen(filename, "rb");
    if(!fd) return (0);
    buf = malloc(size);
    if(!buf) return (0);
    fseek(fd, 0, SEEK_END);
    file_size = ftell(fd);
    if(size != file_size) error("File size mismatch: `%s' is %d bytes, read %d\n", filename, file_size, size);
    fseek(fd, 0, SEEK_SET);
    fread(buf, size, 1, fd);
    fclose(fd);
    for(i = 0; i < size; i += 1)
        mem[i*2+0] = buf[i];
    free(buf);
    return (1);
}

int load_file_odd(char *filename, char *mem, int size)
{
    char *buf;
    int i, file_size;
    FILE *fd = NULL;
    fd = fopen(filename, "rb");
    if(!fd) return (0);
    buf = malloc(size);
    if(!buf) return (0);
    fseek(fd, 0, SEEK_END);
    file_size = ftell(fd);
    if(size != file_size) error("File size mismatch: `%s' is %d bytes, read %d\n", filename, file_size, size);
    fseek(fd, 0, SEEK_SET);
    fread(buf, size, 1, fd);
    fclose(fd);
    for(i = 0; i < size; i += 1)
        mem[i*2+1] = buf[i];
    free(buf);
    return (1);
}


void byteswap(char *mem, int length)
{
    int i, j;
    for(i = 0; i < (length / 2); i += 1)
    {
        j = mem[i*2+0];
        mem[i*2+0] = mem[i*2+1];
        mem[i*2+1] = j;
    }
}


void dos_update_audio(void)
{
    if(option.sound)
    {
        int count;
        signed short int sample[2];

        YM2612UpdateOne(0, snd.fm_buffer, snd.buffer_size);
        SN76496Update(0, snd.psg_buffer, snd.buffer_size);
        UPD7759_update(0, snd.pcm_buffer, snd.buffer_size);

        for(count = 0; count < snd.buffer_size; count += 1)
        {
            sample[0] = sample[1] = 0;
            if(sound_enable & 1) {
            sample[0] += snd.fm_buffer[0][count];
            sample[1] += snd.fm_buffer[1][count];
            }
            if(sound_enable & 2) {
            sample[0] = sample[1] += snd.psg_buffer[count];
            }
            if(sound_enable & 4) {
            sample[0] = sample[1] += snd.pcm_buffer[count];
            }
            snd.buffer[0][count] = sample[0];
            snd.buffer[1][count] = sample[1];
        }

        osd_play_streamed_sample_16(option.swap, &snd.buffer[0][0], snd.buffer_size * 2, option.sndrate, 60, -100);
        osd_play_streamed_sample_16(option.swap ^ 1, &snd.buffer[1][0], snd.buffer_size * 2, option.sndrate, 60,  100);
    }
}


void dos_update_video(void)
{
    int center_x = (SCREEN_W / 2) - (vp_width / 2);
    int center_y = (SCREEN_H / (option.scanlines ? 4 : 2)) - (vp_height / 2);

    /* Clear screen if the resolution changed */
    if(vp_changed) {
        vp_changed = 0;
        clear_to_color(screen, COLOR_BLACK);
    }

    /* Do VSync before blitting and palette updates */
    if(option.vsync) vsync();

    /* Update palette */
    if(palmode == 1)
    {
        int i;
        int bgbasecol = 0x000;
        int sprbasecol = 0x200;
        RGB color;
        word temp;

        if (gab2 == 1)
		{
            switch (palbank2)
            {
                case 0x00:
                    bgbasecol  = 0x000;
                    sprbasecol = 0x200;
                    break;

                case 0x05:
                    bgbasecol  = 0x080;
                    sprbasecol = 0x280;
                    break;

                case 0x0A:
                    bgbasecol  = 0x100;
                    sprbasecol = 0x300;
                    break;

                case 0x0F:
                    bgbasecol  = 0x180;
                    sprbasecol = 0x380;
                    break;
            }
		}

        for(i = 0; i < 0x40; i += 1)
        {
            temp = *(word *)&palram[bgbasecol + (i << 1)];

            /* Normal intensity background palette (00-3F) */
            color.r = ((((temp >> 0) & 0x0F) << 1) | ((temp >> 12) & 1)) << 1;
            color.g = ((((temp >> 4) & 0x0F) << 1) | ((temp >> 13) & 1)) << 1;
            color.b = ((((temp >> 8) & 0x0F) << 1) | ((temp >> 14) & 1)) << 1;
            set_color(0x00 | i, &color);

            /* Low intensity background palette (80-BF) */
            color.r = ((((temp >> 0) & 0x0F) << 1) | ((temp >> 12) & 1));
            color.g = ((((temp >> 4) & 0x0F) << 1) | ((temp >> 13) & 1));
            color.b = ((((temp >> 8) & 0x0F) << 1) | ((temp >> 14) & 1));
            set_color(0x80 | i, &color);

            /* High intensity background palette (C0-FD) */
            color.r = ((((temp >> 0) & 0x0F) << 1) | ((temp >> 12) & 1)) | 0x20;
            color.g = ((((temp >> 4) & 0x0F) << 1) | ((temp >> 13) & 1)) | 0x20;
            color.b = ((((temp >> 8) & 0x0F) << 1) | ((temp >> 14) & 1)) | 0x20;
            if(i < 0x3E)
            set_color(0xC0 | i, &color);
        }

        /* Object palette (40-7F) */
        for(i = 0; i < 0x40; i += 1)
        {
            temp = *(word *)&palram[sprbasecol + (i << 1)];
            color.r = ((((temp >> 0) & 0x0F) << 1) | ((temp >> 12) & 1)) << 1;
            color.g = ((((temp >> 4) & 0x0F) << 1) | ((temp >> 13) & 1)) << 1;
            color.b = ((((temp >> 8) & 0x0F) << 1) | ((temp >> 14) & 1)) << 1;
            set_color(0x40 | i, &color);
        }
    }

    /* Apply blur effect */
    if(option.blur)
        blur(&gen_bmp->line[0][64], 320, 224, 384);

    /* Blit memory bitmap to display according to options used */
    if(option.scanlines)
    {
        int y;
        for(y = 0; y < vp_height; y += 1)
            blit(gen_bmp, screen, BITMAP_MARGIN, y, center_x, (center_y + y) << 1, vp_width, 1);
    }
    else
    {
        if(option.scale)
            stretch_blit(gen_bmp, screen, BITMAP_MARGIN, 0, vp_width, vp_height, 0, 0, SCREEN_W, SCREEN_H);
        else
            blit(gen_bmp, screen, BITMAP_MARGIN, 0, center_x, center_y, vp_width, vp_height);
    }

    /* Do speed throttling */
    if(option.throttle)
    {
        while(tick_count == old_tick_count);
        old_tick_count = tick_count;
    }
}
